#ifndef CAMLIBS_LARGAN_LMINI_LMINI_CCD_H
#define CAMLIBS_LARGAN_LMINI_LMINI_CCD_H

void largan_ccd2dib(char *pData, char *pDib, long dwDibRowBytes, int nCcdFactor);

#endif /* !defined(CAMLIBS_LARGAN_LMINI_LMINI_CCD_H) */

